"""WeDo sensor handling functionality"""

from .devices import Device


class TiltSensor(Device):
    """Tilt sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no tilt sensor attached to port
    """

    def __init__(self, port):
        """
        Initialise tilt sensor

        :param port: Port of device
        """
        super().__init__(port)
        self.mode(0)

    def get_tilt(self):
        """
        Return the tilt from tilt sensor

        :return: Tilt from tilt sensor
        :rtype: tuple
        """
        return tuple(self.get())


class MotionSensor(Device):
    """Motion sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no motion sensor attached to port
    """

    default_mode = 0

    def __init__(self, port):
        """
        Initialise motion sensor

        :param port: Port of device
        """
        super().__init__(port)
        self.mode(self.default_mode)

    def set_default_data_mode(self, mode):
        """
        Set the mode most often queried from this device.

        This significantly improves performance when repeatedly accessing data

        :param mode: 0 for distance (default), 1 for movement count
        """
        if mode == 1 or mode == 0:
            self.default_mode = mode
            self.mode(mode)

    def get_distance(self):
        """
        Return the distance from motion sensor

        :return: Distance from motion sensor
        :rtype: int
        """
        return self._get_data_from_mode(0)

    def get_movement_count(self):
        """
        Return the movement counter

        This is the count of how many times the sensor has detected an object
        that moved within 4 blocks of the sensor since the sensor has been
        plugged in or the BuildHAT reset

        :return: Count of objects detected
        :rtype: int
        """
        return self._get_data_from_mode(1)

    def _get_data_from_mode(self, mode):
        if self.default_mode == mode:
            return self.get()[0]
        else:
            self.mode(mode)
            retval = self.get()[0]
            self.mode(self.default_mode)
            return retval



"""Motor device handling functionality"""

import threading
import time
from collections import deque
from concurrent.futures import Future
from enum import Enum
from threading import Condition

from .devices import Device
from .exc import MotorError


class PassiveMotor(Device):
    """Passive Motor device

    :param port: Port of device
    :raises DeviceError: Occurs if there is no passive motor attached to port
    """

    def __init__(self, port):
        """Initialise motor

        :param port: Port of device
        """
        super().__init__(port)
        self._default_speed = 20
        self._currentspeed = 0
        self.plimit(0.7)

    def set_default_speed(self, default_speed):
        """Set the default speed of the motor

        :param default_speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs if invalid speed passed
        """
        if not (default_speed >= -100 and default_speed <= 100):
            raise MotorError("Invalid Speed")
        self._default_speed = default_speed

    def start(self, speed=None):
        """Start motor

        :param speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs if invalid speed passed
        """
        if self._currentspeed == speed:
            # Already running at this speed, do nothing
            return

        if speed is None:
            speed = self._default_speed
        else:
            if not (speed >= -100 and speed <= 100):
                raise MotorError("Invalid Speed")
        self._currentspeed = speed
        cmd = f"port {self.port} ; pwm ; set {speed / 100}\r"
        self._write(cmd)

    def stop(self):
        """Stop motor"""
        cmd = f"port {self.port} ; off\r"
        self._write(cmd)
        self._currentspeed = 0

    def plimit(self, plimit):
        """Limit power

        :param plimit: Value 0 to 1
        :raises MotorError: Occurs if invalid plimit value passed
        """
        if not (plimit >= 0 and plimit <= 1):
            raise MotorError("plimit should be 0 to 1")
        self._write(f"port {self.port} ; port_plimit {plimit}\r")

    def bias(self, bias):
        """Bias motor

        :param bias: Value 0 to 1
        :raises MotorError: Occurs if invalid bias value passed

        .. deprecated:: 0.6.0
        """  # noqa: RST303
        raise MotorError("Bias no longer available")


class MotorRunmode(Enum):
    """Current mode motor is in"""

    NONE = 0
    FREE = 1
    DEGREES = 2
    SECONDS = 3


class Motor(Device):
    """Motor device

    :param port: Port of device
    :raises DeviceError: Occurs if there is no motor attached to port
    """

    def __init__(self, port):
        """Initialise motor

        :param port: Port of device
        """
        super().__init__(port)
        self.default_speed = 20
        self._currentspeed = 0
        if self._typeid in {38}:
            self.mode([(1, 0), (2, 0)])
            self._combi = "1 0 2 0"
            self._noapos = True
        else:
            self.mode([(1, 0), (2, 0), (3, 0)])
            self._combi = "1 0 2 0 3 0"
            self._noapos = False
        self.plimit(0.7)
        self.pwmparams(0.65, 0.01)
        self._rpm = False
        self._release = True
        self._bqueue = deque(maxlen=5)
        self._cvqueue = Condition()
        self.when_rotated = None
        self._oldpos = None
        self._runmode = MotorRunmode.NONE

    def set_speed_unit_rpm(self, rpm=False):
        """Set whether to use RPM for speed units or not

        :param rpm: Boolean to determine whether to use RPM for units
        """
        self._rpm = rpm

    def set_default_speed(self, default_speed):
        """Set the default speed of the motor

        :param default_speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs if invalid speed passed
        """
        if not (default_speed >= -100 and default_speed <= 100):
            raise MotorError("Invalid Speed")
        self.default_speed = default_speed

    def run_for_rotations(self, rotations, speed=None, blocking=True):
        """Run motor for N rotations

        :param rotations: Number of rotations
        :param speed: Speed ranging from -100 to 100
        :param blocking: Whether call should block till finished
        :raises MotorError: Occurs if invalid speed passed
        """
        self._runmode = MotorRunmode.DEGREES
        if speed is None:
            self.run_for_degrees(int(rotations * 360), self.default_speed, blocking)
        else:
            if not (speed >= -100 and speed <= 100):
                raise MotorError("Invalid Speed")
            self.run_for_degrees(int(rotations * 360), speed, blocking)

    def _run_for_degrees(self, degrees, speed):
        self._runmode = MotorRunmode.DEGREES
        mul = 1
        if speed < 0:
            speed = abs(speed)
            mul = -1
        pos = self.get_position()
        newpos = ((degrees * mul) + pos) / 360.0
        pos /= 360.0
        self._run_positional_ramp(pos, newpos, speed)
        self._runmode = MotorRunmode.NONE

    def _run_to_position(self, degrees, speed, direction):
        self._runmode = MotorRunmode.DEGREES
        data = self.get()
        pos = data[1]
        if self._noapos:
            apos = pos
        else:
            apos = data[2]
        diff = (degrees - apos + 180) % 360 - 180
        newpos = (pos + diff) / 360
        v1 = (degrees - apos) % 360
        v2 = (apos - degrees) % 360
        mul = 1
        if diff > 0:
            mul = -1
        diff = sorted([diff, mul * (v2 if abs(diff) == v1 else v1)])
        if direction == "shortest":
            pass
        elif direction == "clockwise":
            newpos = (pos + diff[1]) / 360
        elif direction == "anticlockwise":
            newpos = (pos + diff[0]) / 360
        else:
            raise MotorError("Invalid direction, should be: shortest, clockwise or anticlockwise")
        # Convert current motor position to decimal rotations from preset position to match newpos units
        pos /= 360.0
        self._run_positional_ramp(pos, newpos, speed)
        self._runmode = MotorRunmode.NONE

    def _run_positional_ramp(self, pos, newpos, speed):
        """Ramp motor

        :param pos: Current motor position in decimal rotations (from preset position)
        :param newpos: New motor postion in decimal rotations (from preset position)
        :param speed: -100 to 100
        """
        if self._rpm:
            speed = self._speed_process(speed)
        else:
            speed *= 0.05  # Collapse speed range to -5 to 5
        dur = abs((newpos - pos) / speed)
        cmd = (f"port {self.port}; select 0 ; selrate {self._interval}; "
               f"pid {self.port} 0 1 s4 0.0027777778 0 5 0 .1 3 0.01; "
               f"set ramp {pos} {newpos} {dur} 0\r")
        ftr = Future()
        self._hat.rampftr[self.port].append(ftr)
        self._write(cmd)
        ftr.result()
        if self._release:
            time.sleep(0.2)
            self.coast()

    def run_for_degrees(self, degrees, speed=None, blocking=True):
        """Run motor for N degrees

        Speed of 1 means 1 revolution / second

        :param degrees: Number of degrees to rotate
        :param speed: Speed ranging from -100 to 100
        :param blocking: Whether call should block till finished
        :raises MotorError: Occurs if invalid speed passed
        """
        self._runmode = MotorRunmode.DEGREES
        if speed is None:
            speed = self.default_speed
        if not (speed >= -100 and speed <= 100):
            raise MotorError("Invalid Speed")
        if not blocking:
            self._queue((self._run_for_degrees, (degrees, speed)))
        else:
            self._wait_for_nonblocking()
            self._run_for_degrees(degrees, speed)

    def run_to_position(self, degrees, speed=None, blocking=True, direction="shortest"):
        """Run motor to position (in degrees)

        :param degrees: Position in degrees from -180 to 180
        :param speed: Speed ranging from 0 to 100
        :param blocking: Whether call should block till finished
        :param direction: shortest (default)/clockwise/anticlockwise
        :raises MotorError: Occurs if invalid speed or angle passed
        """
        self._runmode = MotorRunmode.DEGREES
        if speed is None:
            speed = self.default_speed
        if not (speed >= 0 and speed <= 100):
            raise MotorError("Invalid Speed")
        if degrees < -180 or degrees > 180:
            raise MotorError("Invalid angle")
        if not blocking:
            self._queue((self._run_to_position, (degrees, speed, direction)))
        else:
            self._wait_for_nonblocking()
            self._run_to_position(degrees, speed, direction)

    def _run_for_seconds(self, seconds, speed):
        speed = self._speed_process(speed)
        self._runmode = MotorRunmode.SECONDS
        if self._rpm:
            pid = f"pid_diff {self.port} 0 5 s2 0.0027777778 1 0 2.5 0 .4 0.01; "
        else:
            pid = f"pid {self.port} 0 0 s1 1 0 0.003 0.01 0 100 0.01;"
        cmd = (f"port {self.port} ; select 0 ; selrate {self._interval}; "
               f"{pid}"
               f"set pulse {speed} 0.0 {seconds} 0\r")
        ftr = Future()
        self._hat.pulseftr[self.port].append(ftr)
        self._write(cmd)
        ftr.result()
        if self._release:
            self.coast()
        self._runmode = MotorRunmode.NONE

    def run_for_seconds(self, seconds, speed=None, blocking=True):
        """Run motor for N seconds

        :param seconds: Time in seconds
        :param speed: Speed ranging from -100 to 100
        :param blocking: Whether call should block till finished
        :raises MotorError: Occurs when invalid speed specified
        """
        self._runmode = MotorRunmode.SECONDS
        if speed is None:
            speed = self.default_speed
        if not (speed >= -100 and speed <= 100):
            raise MotorError("Invalid Speed")
        if not blocking:
            self._queue((self._run_for_seconds, (seconds, speed)))
        else:
            self._wait_for_nonblocking()
            self._run_for_seconds(seconds, speed)

    def start(self, speed=None):
        """Start motor

        :param speed: Speed ranging from -100 to 100
        :raises MotorError: Occurs when invalid speed specified
        """
        self._wait_for_nonblocking()
        if self._runmode == MotorRunmode.FREE:
            if self._currentspeed == speed:
                # Already running at this speed, do nothing
                return
        elif self._runmode != MotorRunmode.NONE:
            # Motor is running some other mode, wait for it to stop or stop() it yourself
            return

        if speed is None:
            speed = self.default_speed
        else:
            if not (speed >= -100 and speed <= 100):
                raise MotorError("Invalid Speed")
        speed = self._speed_process(speed)
        cmd = f"port {self.port} ; set {speed}\r"
        if self._runmode == MotorRunmode.NONE:
            if self._rpm:
                pid = f"pid_diff {self.port} 0 5 s2 0.0027777778 1 0 2.5 0 .4 0.01; "
            else:
                pid = f"pid {self.port} 0 0 s1 1 0 0.003 0.01 0 100 0.01; "
            cmd = (f"port {self.port} ; select 0 ; selrate {self._interval}; "
                   f"{pid}"
                   f"set {speed}\r")
        self._runmode = MotorRunmode.FREE
        self._currentspeed = speed
        self._write(cmd)

    def stop(self):
        """Stop motor"""
        self._wait_for_nonblocking()
        self._runmode = MotorRunmode.NONE
        self._currentspeed = 0
        self.coast()

    def get_position(self):
        """Get position of motor with relation to preset position (can be negative or positive)

        :return: Position of motor in degrees from preset position
        :rtype: int
        """
        return self.get()[1]

    def get_aposition(self):
        """Get absolute position of motor

        :return: Absolute position of motor from -180 to 180
        :rtype: int
        """
        if self._noapos:
            raise MotorError("No absolute position with this motor")
        else:
            return self.get()[2]

    def get_speed(self):
        """Get speed of motor

        :return: Speed of motor
        :rtype: int
        """
        return self.get()[0]

    @property
    def when_rotated(self):
        """
        Handle rotation events

        :getter: Returns function to be called when rotated
        :setter: Sets function to be called when rotated
        :return: Callback function
        """
        return self._when_rotated

    def _intermediate(self, data):
        if self._noapos:
            speed, pos = data
            apos = None
        else:
            speed, pos, apos = data
        if self._oldpos is None:
            self._oldpos = pos
            return
        if abs(pos - self._oldpos) >= 1:
            if self._when_rotated is not None:
                self._when_rotated(speed, pos, apos)
            self._oldpos = pos

    @when_rotated.setter
    def when_rotated(self, value):
        """Call back, when motor has been rotated

        :param value: Callback function
        """
        self._when_rotated = value
        self.callback(self._intermediate)

    def plimit(self, plimit):
        """Limit power

        :param plimit: Value 0 to 1
        :raises MotorError: Occurs if invalid plimit value passed
        """
        if not (plimit >= 0 and plimit <= 1):
            raise MotorError("plimit should be 0 to 1")
        self._write(f"port {self.port} ; port_plimit {plimit}\r")

    def bias(self, bias):
        """Bias motor

        :param bias: Value 0 to 1
        :raises MotorError: Occurs if invalid bias value passed

        .. deprecated:: 0.6.0
        """  # noqa: RST303
        raise MotorError("Bias no longer available")

    def pwmparams(self, pwmthresh, minpwm):
        """PWM thresholds

        :param pwmthresh: Value 0 to 1, threshold below, will switch from fast to slow, PWM
        :param minpwm: Value 0 to 1, threshold below which it switches off the drive altogether
        :raises MotorError: Occurs if invalid values are passed
        """
        if not (pwmthresh >= 0 and pwmthresh <= 1):
            raise MotorError("pwmthresh should be 0 to 1")
        if not (minpwm >= 0 and minpwm <= 1):
            raise MotorError("minpwm should be 0 to 1")
        self._write(f"port {self.port} ; pwmparams {pwmthresh} {minpwm}\r")

    def pwm(self, pwmv):
        """PWM motor

        :param pwmv: Value -1 to 1
        :raises MotorError: Occurs if invalid pwm value passed
        """
        if not (pwmv >= -1 and pwmv <= 1):
            raise MotorError("pwm should be -1 to 1")
        self._write(f"port {self.port} ; pwm ; set {pwmv}\r")

    def coast(self):
        """Coast motor"""
        self._write(f"port {self.port} ; coast\r")

    def float(self):
        """Float motor"""
        self.pwm(0)

    @property
    def release(self):
        """Determine if motor is released after running, so can be turned by hand

        :getter: Returns whether motor is released, so can be turned by hand
        :setter: Sets whether motor is released, so can be turned by hand
        :return: Whether motor is released, so can be turned by hand
        :rtype: bool
        """
        return self._release

    @release.setter
    def release(self, value):
        """Determine if the motor is released after running, so can be turned by hand

        :param value: Whether motor should be released, so can be turned by hand
        :type value: bool
        """
        if not isinstance(value, bool):
            raise MotorError("Must pass boolean")
        self._release = value

    def _queue(self, cmd):
        Device._instance.motorqueue[self.port].put(cmd)

    def _wait_for_nonblocking(self):
        """Wait for nonblocking commands to finish"""
        Device._instance.motorqueue[self.port].join()

    def _speed_process(self, speed):
        """Lower speed value"""
        if self._rpm:
            return speed / 60
        else:
            return speed


class MotorPair:
    """Pair of motors

    :param motora: One of the motors to drive
    :param motorb: Other motor in pair to drive
    :raises DeviceError: Occurs if there is no motor attached to port
    """

    def __init__(self, leftport, rightport):
        """Initialise pair of motors

        :param leftport: Left motor port
        :param rightport:  Right motor port
        """
        super().__init__()
        self._leftmotor = Motor(leftport)
        self._rightmotor = Motor(rightport)
        self.default_speed = 20
        self._release = True
        self._rpm = False

    def set_default_speed(self, default_speed):
        """Set the default speed of the motor

        :param default_speed: Speed ranging from -100 to 100
        """
        self.default_speed = default_speed

    def set_speed_unit_rpm(self, rpm=False):
        """Set whether to use RPM for speed units or not

        :param rpm: Boolean to determine whether to use RPM for units
        """
        self._rpm = rpm
        self._leftmotor.set_speed_unit_rpm(rpm)
        self._rightmotor.set_speed_unit_rpm(rpm)

    def run_for_rotations(self, rotations, speedl=None, speedr=None):
        """Run pair of motors for N rotations

        :param rotations: Number of rotations
        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        self.run_for_degrees(int(rotations * 360), speedl, speedr)

    def run_for_degrees(self, degrees, speedl=None, speedr=None):
        """Run pair of motors for degrees

        :param degrees: Number of degrees
        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        th1 = threading.Thread(target=self._leftmotor._run_for_degrees, args=(degrees, speedl))
        th2 = threading.Thread(target=self._rightmotor._run_for_degrees, args=(degrees, speedr))
        th1.daemon = True
        th2.daemon = True
        th1.start()
        th2.start()
        th1.join()
        th2.join()

    def run_for_seconds(self, seconds, speedl=None, speedr=None):
        """Run pair for N seconds

        :param seconds: Time in seconds
        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        th1 = threading.Thread(target=self._leftmotor._run_for_seconds, args=(seconds, speedl))
        th2 = threading.Thread(target=self._rightmotor._run_for_seconds, args=(seconds, speedr))
        th1.daemon = True
        th2.daemon = True
        th1.start()
        th2.start()
        th1.join()
        th2.join()

    def start(self, speedl=None, speedr=None):
        """Start motors

        :param speedl: Speed ranging from -100 to 100
        :param speedr: Speed ranging from -100 to 100
        """
        if speedl is None:
            speedl = self.default_speed
        if speedr is None:
            speedr = self.default_speed
        self._leftmotor.start(speedl)
        self._rightmotor.start(speedr)

    def stop(self):
        """Stop motors"""
        self._leftmotor.stop()
        self._rightmotor.stop()

    def run_to_position(self, degreesl, degreesr, speed=None, direction="shortest"):
        """Run pair to position (in degrees)

        :param degreesl: Position in degrees for left motor
        :param degreesr: Position in degrees for right motor
        :param speed: Speed ranging from -100 to 100
        :param direction: shortest (default)/clockwise/anticlockwise
        """
        if speed is None:
            th1 = threading.Thread(target=self._leftmotor._run_to_position, args=(degreesl, self.default_speed, direction))
            th2 = threading.Thread(target=self._rightmotor._run_to_position, args=(degreesr, self.default_speed, direction))
        else:
            th1 = threading.Thread(target=self._leftmotor._run_to_position, args=(degreesl, speed, direction))
            th2 = threading.Thread(target=self._rightmotor._run_to_position, args=(degreesr, speed, direction))
        th1.daemon = True
        th2.daemon = True
        th1.start()
        th2.start()
        th1.join()
        th2.join()

    @property
    def release(self):
        """Determine if motors are released after running, so can be turned by hand

        :getter: Returns whether motors are released, so can be turned by hand
        :setter: Sets whether motors are released, so can be turned by hand
        :return: Whether motors are released, so can be turned by hand
        :rtype: bool
        """
        return self._release

    @release.setter
    def release(self, value):
        """Determine if motors are released after running, so can be turned by hand

        :param value: Whether motors should be released, so can be turned by hand
        :type value: bool
        """
        if not isinstance(value, bool):
            raise MotorError("Must pass boolean")
        self._release = value
        self._leftmotor.release = value
        self._rightmotor.release = value



"""Color distance sensor handling functionality"""

import math
from collections import deque
from threading import Condition

from .devices import Device


class ColorDistanceSensor(Device):
    """Color Distance sensor

    :param port: Port of device
    :raises DeviceError: Occurs if there is no colordistance sensor attached to port
    """

    def __init__(self, port):
        """
        Initialise color distance sensor

        :param port: Port of device
        """
        super().__init__(port)
        self.on()
        self.mode(6)
        self.avg_reads = 4
        self._old_color = None
        self._ir_channel = 0x0
        self._ir_address = 0x0
        self._ir_toggle = 0x0

    def segment_color(self, r, g, b):
        """Return the color name from HSV

        :param r: Red
        :param g: Green
        :param b: Blue
        :return: Name of the color as a string
        :rtype: str
        """
        table = [("black", (0, 0, 0)),
                 ("violet", (127, 0, 255)),
                 ("blue", (0, 0, 255)),
                 ("cyan", (0, 183, 235)),
                 ("green", (0, 128, 0)),
                 ("yellow", (255, 255, 0)),
                 ("red", (255, 0, 0)),
                 ("white", (255, 255, 255))]
        near = ""
        euc = math.inf
        for itm in table:
            cur = math.sqrt((r - itm[1][0]) ** 2 + (g - itm[1][1]) ** 2 + (b - itm[1][2]) ** 2)
            if cur < euc:
                near = itm[0]
                euc = cur
        return near

    def rgb_to_hsv(self, r, g, b):
        """Convert RGB to HSV

        Based on https://www.rapidtables.com/convert/color/rgb-to-hsv.html algorithm

        :param r: Red
        :param g: Green
        :param b: Blue
        :return: HSV representation of color
        :rtype: tuple
        """
        r, g, b = r / 255.0, g / 255.0, b / 255.0
        cmax = max(r, g, b)
        cmin = min(r, g, b)
        delt = cmax - cmin
        if cmax == cmin:
            h = 0
        elif cmax == r:
            h = 60 * (((g - b) / delt) % 6)
        elif cmax == g:
            h = 60 * ((((b - r) / delt)) + 2)
        elif cmax == b:
            h = 60 * ((((r - g) / delt)) + 4)
        if cmax == 0:
            s = 0
        else:
            s = delt / cmax
        v = cmax
        return int(h), int(s * 100), int(v * 100)

    def get_color(self):
        """Return the color

        :return: Name of the color as a string
        :rtype: str
        """
        r, g, b = self.get_color_rgb()
        return self.segment_color(r, g, b)

    def get_ambient_light(self):
        """Return the ambient light

        :return: Ambient light
        :rtype: int
        """
        self.mode(4)
        readings = []
        for _ in range(self.avg_reads):
            readings.append(self.get()[0])
        return int(sum(readings) / len(readings))

    def get_reflected_light(self):
        """Return the reflected light

        :return: Reflected light
        :rtype: int
        """
        self.mode(3)
        readings = []
        for _ in range(self.avg_reads):
            readings.append(self.get()[0])
        return int(sum(readings) / len(readings))

    def get_distance(self):
        """Return the distance

        :return: Distance
        :rtype: int
        """
        self.mode(1)
        distance = self.get()[0]
        return distance

    def _clamp(self, val, small, large):
        return max(small, min(val, large))

    def _avgrgb(self, reads):
        readings = []
        for read in reads:
            read = [int((self._clamp(read[0], 0, 400) / 400) * 255),
                    int((self._clamp(read[1], 0, 400) / 400) * 255),
                    int((self._clamp(read[2], 0, 400) / 400) * 255)]
            readings.append(read)
        rgb = []
        for i in range(3):
            rgb.append(int(sum([rgb[i] for rgb in readings]) / len(readings)))
        return rgb

    def get_color_rgb(self):
        """Return the color

        :return: RGBI representation
        :rtype: list
        """
        self.mode(6)
        reads = []
        for _ in range(self.avg_reads):
            reads.append(self.get())
        return self._avgrgb(reads)

    def _cb_handle(self, lst):
        self._data.append(lst)
        if len(self._data) == self.avg_reads:
            r, g, b = self._avgrgb(self._data)
            seg = self.segment_color(r, g, b)
            if self._cmp(seg, self._color):
                with self._cond:
                    self._old_color = seg
                    self._cond.notify()

    def wait_until_color(self, color):
        """Wait until specific color

        :param color: Color to look for
        """
        self.mode(6)
        self._cond = Condition()
        self._data = deque(maxlen=self.avg_reads)
        self._color = color
        self._cmp = lambda x, y: x == y
        self.callback(self._cb_handle)
        with self._cond:
            self._cond.wait()
        self.callback(None)

    def wait_for_new_color(self):
        """Wait for new color or returns immediately if first call

        :return: Name of the color as a string
        :rtype: str
        """
        self.mode(6)
        if self._old_color is None:
            self._old_color = self.get_color()
            return self._old_color
        self._cond = Condition()
        self._data = deque(maxlen=self.avg_reads)
        self._color = self._old_color
        self._cmp = lambda x, y: x != y
        self.callback(self._cb_handle)
        with self._cond:
            self._cond.wait()
        self.callback(None)
        return self._old_color

    @property
    def ir_channel(self):
        """Get the IR channel for message transmission"""
        return self._ir_channel

    @ir_channel.setter
    def ir_channel(self, channel=1):
        """
        Set the IR channel for RC Tx

        :param channel: 1-4 indicating the selected IR channel on the reciever
        """
        check_chan = channel
        if check_chan > 4:
            check_chan = 4
        elif check_chan < 1:
            check_chan = 1
        # Internally: 0-3
        self._ir_channel = int(check_chan) - 1

    @property
    def ir_address(self):
        """IR Address space of 0x0 for default PoweredUp or 0x1 for extra space"""
        return self._ir_address

    def toggle_ir_toggle(self):
        """Toggle the IR toggle bit"""
        # IYKYK, because the RC documents are not clear
        if self._ir_toggle:
            self._ir_toggle = 0x0
        else:
            self._ir_toggle = 0x1
        return self._ir_toggle

    def send_ir_sop(self, port, mode):
        """
        Send an IR message via Power Functions RC Protocol in Single Output PWM mode

        PF IR RC Protocol documented at https://www.philohome.com/pf/pf.htm

        Port B is blue

        Valid values for mode are:
        0x0: Float output
        0x1: Forward/Clockwise at speed 1
        0x2: Forward/Clockwise at speed 2
        0x3: Forward/Clockwise at speed 3
        0x4: Forward/Clockwise at speed 4
        0x5: Forward/Clockwise at speed 5
        0x6: Forward/Clockwise at speed 6
        0x7: Forward/Clockwise at speed 7
        0x8: Brake (then float v1.20)
        0x9: Backwards/Counterclockwise at speed 7
        0xA: Backwards/Counterclockwise at speed 6
        0xB: Backwards/Counterclockwise at speed 5
        0xC: Backwards/Counterclockwise at speed 4
        0xD: Backwards/Counterclockwise at speed 3
        0xE: Backwards/Counterclockwise at speed 2
        0xF: Backwards/Counterclockwise at speed 1

        :param port: 'A' or 'B'
        :param mode: 0-15 indicating the port's mode to set
        """
        escape_modeselect = 0x0
        escape = escape_modeselect

        ir_mode_single_output = 0x4
        ir_mode = ir_mode_single_output

        so_mode_pwm = 0x0
        so_mode = so_mode_pwm

        output_port_a = 0x0
        output_port_b = 0x1

        output_port = None
        if port == 'A' or port == 'a':
            output_port = output_port_a
        elif port == 'B' or port == 'b':
            output_port = output_port_b
        else:
            return False

        ir_mode = ir_mode | (so_mode << 1) | output_port

        nibble1 = (self._ir_toggle << 3) | (escape << 2) | self._ir_channel
        nibble2 = (self._ir_address << 3) | ir_mode

        # Mode range checked here
        return self._send_ir_nibbles(nibble1, nibble2, mode)

    def send_ir_socstid(self, port, mode):
        """
        Send an IR message via Power Functions RC Protocol in Single Output Clear/Set/Toggle/Increment/Decrement mode

        PF IR RC Protocol documented at https://www.philohome.com/pf/pf.htm

        Valid values for mode are:
        0x0: Toggle full Clockwise/Forward (Stop to Clockwise, Clockwise to Stop, Counterclockwise to Clockwise)
        0x1: Toggle direction
        0x2: Increment numerical PWM
        0x3: Decrement numerical PWM
        0x4: Increment PWM
        0x5: Decrement PWM
        0x6: Full Clockwise/Forward
        0x7: Full Counterclockwise/Backward
        0x8: Toggle full (defaults to Forward, first)
        0x9: Clear C1 (C1 to High)
        0xA: Set C1 (C1 to Low)
        0xB: Toggle C1
        0xC: Clear C2 (C2 to High)
        0xD: Set C2 (C2 to Low)
        0xE: Toggle C2
        0xF: Toggle full Counterclockwise/Backward (Stop to Clockwise, Counterclockwise to Stop, Clockwise to Counterclockwise)

        :param port: 'A' or 'B'
        :param mode: 0-15 indicating the port's mode to set
        """
        escape_modeselect = 0x0
        escape = escape_modeselect

        ir_mode_single_output = 0x4
        ir_mode = ir_mode_single_output

        so_mode_cstid = 0x1
        so_mode = so_mode_cstid

        output_port_a = 0x0
        output_port_b = 0x1

        output_port = None
        if port == 'A' or port == 'a':
            output_port = output_port_a
        elif port == 'B' or port == 'b':
            output_port = output_port_b
        else:
            return False

        ir_mode = ir_mode | (so_mode << 1) | output_port

        nibble1 = (self._ir_toggle << 3) | (escape << 2) | self._ir_channel
        nibble2 = (self._ir_address << 3) | ir_mode

        # Mode range checked here
        return self._send_ir_nibbles(nibble1, nibble2, mode)

    def send_ir_combo_pwm(self, port_b_mode, port_a_mode):
        """
        Send an IR message via Power Functions RC Protocol in Combo PWM mode

        PF IR RC Protocol documented at https://www.philohome.com/pf/pf.htm

        Valid values for the modes are:
        0x0 Float
        0x1 PWM Forward step 1
        0x2 PWM Forward step 2
        0x3 PWM Forward step 3
        0x4 PWM Forward step 4
        0x5 PWM Forward step 5
        0x6 PWM Forward step 6
        0x7 PWM Forward step 7
        0x8 Brake (then float v1.20)
        0x9 PWM Backward step 7
        0xA PWM Backward step 6
        0xB PWM Backward step 5
        0xC PWM Backward step 4
        0xD PWM Backward step 3
        0xE PWM Backward step 2
        0xF PWM Backward step 1

        :param port_b_mode: 0-15 indicating the command to send to port B
        :param port_a_mode: 0-15 indicating the command to send to port A
        """
        escape_combo_pwm = 0x1
        escape = escape_combo_pwm

        nibble1 = (self._ir_toggle << 3) | (escape << 2) | self._ir_channel

        # Port modes are range checked here
        return self._send_ir_nibbles(nibble1, port_b_mode, port_a_mode)

    def send_ir_combo_direct(self, port_b_output, port_a_output):
        """
        Send an IR message via Power Functions RC Protocol in Combo Direct mode

        PF IR RC Protocol documented at https://www.philohome.com/pf/pf.htm

        Valid values for the output variables are:
        0x0: Float output
        0x1: Clockwise/Forward
        0x2: Counterclockwise/Backwards
        0x3: Brake then float

        :param port_b_output: 0-3 indicating the output to send to port B
        :param port_a_output: 0-3 indicating the output to send to port A
        """
        escape_modeselect = 0x0
        escape = escape_modeselect

        ir_mode_combo_direct = 0x1
        ir_mode = ir_mode_combo_direct

        nibble1 = (self._ir_toggle << 3) | (escape << 2) | self._ir_channel
        nibble2 = (self._ir_address << 3) | ir_mode

        if port_b_output > 0x3 or port_a_output > 0x3:
            return False
        if port_b_output < 0x0 or port_a_output < 0x0:
            return False

        nibble3 = (port_b_output << 2) | port_a_output

        return self._send_ir_nibbles(nibble1, nibble2, nibble3)

    def send_ir_extended(self, mode):
        """
        Send an IR message via Power Functions RC Protocol in Extended mode

        PF IR RC Protocol documented at https://www.philohome.com/pf/pf.htm

        Valid values for the mode are:
        0x0: Brake Port A (timeout)
        0x1: Increment Speed on Port A
        0x2: Decrement Speed on Port A

        0x4: Toggle Forward/Clockwise/Float on Port B

        0x6: Toggle Address bit
        0x7: Align toggle bit

        :param mode: 0-2,4,6-7
        """
        escape_modeselect = 0x0
        escape = escape_modeselect

        ir_mode_extended = 0x0
        ir_mode = ir_mode_extended

        nibble1 = (self._ir_toggle << 3) | (escape << 2) | self._ir_channel
        nibble2 = (self._ir_address << 3) | ir_mode

        if mode < 0x0 or mode == 0x3 or mode == 0x5 or mode > 0x7:
            return False

        return self._send_ir_nibbles(nibble1, nibble2, mode)

    def send_ir_single_pin(self, port, pin, mode, timeout):
        """
        Send an IR message via Power Functions RC Protocol in Single Pin mode

        PF IR RC Protocol documented at https://www.philohome.com/pf/pf.htm

        Valid values for the mode are:
        0x0: No-op
        0x1: Clear
        0x2: Set
        0x3: Toggle

        Note: The unlabeled IR receiver (vs the one labeled V2) has a "firmware bug in Single Pin mode"
        https://www.philohome.com/pfrec/pfrec.htm

        :param port: 'A' or 'B'
        :param pin: 1 or 2
        :param mode: 0-3 indicating the pin's mode to set
        :param timeout: True or False
        """
        escape_mode = 0x0
        escape = escape_mode

        ir_mode_single_continuous = 0x2
        ir_mode_single_timeout = 0x3
        ir_mode = None
        if timeout:
            ir_mode = ir_mode_single_timeout
        else:
            ir_mode = ir_mode_single_continuous

        output_port_a = 0x0
        output_port_b = 0x1

        output_port = None
        if port == 'A' or port == 'a':
            output_port = output_port_a
        elif port == 'B' or port == 'b':
            output_port = output_port_b
        else:
            return False

        if pin != 1 and pin != 2:
            return False
        pin_value = pin - 1

        if mode > 0x3 or mode < 0x0:
            return False

        nibble1 = (self._ir_toggle << 3) | (escape << 2) | self._ir_channel
        nibble2 = (self._ir_address << 3) | ir_mode
        nibble3 = (output_port << 3) | (pin_value << 3) | mode

        return self._send_ir_nibbles(nibble1, nibble2, nibble3)

    def _send_ir_nibbles(self, nibble1, nibble2, nibble3):

        #  M7 IR Tx SI = N/A
        #    format count=1 type=1 chars=5 dp=0
        #    RAW: 00000000 0000FFFF    PCT: 00000000 00000064    SI: 00000000 0000FFFF

        mode = 7
        self.mode(mode)

        # The upper bits of data[2] are ignored
        if nibble1 > 0xF or nibble2 > 0xF or nibble3 > 0xF:
            return False
        if nibble1 < 0x0 or nibble2 < 0x0 or nibble3 < 0x0:
            return False

        byte_two = (nibble2 << 4) | nibble3

        data = bytearray(3)
        data[0] = (0xc << 4) | mode
        data[1] = byte_two
        data[2] = nibble1

        # print(" ".join('{:04b}'.format(nibble1)))
        # print(" ".join('{:04b}'.format(nibble2)))
        # print(" ".join('{:04b}'.format(nibble3)))
        # print(" ".join('{:08b}'.format(n) for n in data))

        self._write1(data)
        return True

    def on(self):
        """Turn on the sensor and LED"""
        self.reverse()
